from .base import ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin


class Products(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "products"


class JournalEntries(CreateAPIMixin):
    collection = "journal_entries"


class Invoices(ListAPIMixin, GetAPIMixin):
    collection = "invoices"


class Customers(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "customers"


class Payments(ListAPIMixin, GetAPIMixin, CreateAPIMixin):
    collection = "payments"


class PaymentRuns(GetAPIMixin, CreateAPIMixin):
    collection = "payment_runs"


class Credits(ListAPIMixin, GetAPIMixin, CreateAPIMixin):
    collection = "credits"


class Refunds(ListAPIMixin, GetAPIMixin):
    collection = "refunds"


class Statements(ListAPIMixin, GetAPIMixin):
    collection = "statements"


class Plans(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "plans"


class Coupons(ListAPIMixin, GetAPIMixin, CreateAPIMixin):
    collection = "coupons"


class Subscriptions(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "subscriptions"


class Orders(ListAPIMixin, GetAPIMixin, CreateAPIMixin):
    collection = "orders"


class Usages(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "usages"


class Webhooks(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "webhooks"


class BillingRuns(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "billing_runs"


class RevenueSchedules(ListAPIMixin, GetAPIMixin):
    MAX_PAGE_SIZE = 500

    collection = "revenue_schedules"


class BillingSchedules(ListAPIMixin, GetAPIMixin):
    collection = "billing_schedules"


class RevenueRules(ListAPIMixin, GetAPIMixin, CreateAPIMixin, DeleteAPIMixin):
    collection = "revenue_rules"


class ChartOfAccounts(ListAPIMixin, GetAPIMixin, CreateAPIMixin):
    collection = "chart_of_accounts"



from typing import Optional, Dict, Union
from ordway.exceptions import OrdwayClientException
from requests.exceptions import RequestException


class OrdwayAPIException(OrdwayClientException):
    """ Base exception for all Ordway API errors. """


class OrdwayAPIRequestException(OrdwayAPIException, RequestException):
    def __init__(self, *args, **kwargs):
        """ Initalizes an OrdwayAPIRequestException with `errors` object easily accessible, if returned. """

        self.errors: Optional[
            Dict[str, Union[int, str, Dict[str, Dict[str, str], str]]]
        ] = kwargs.pop("errors", None)

        if self.errors is None:
            # Attempt to grab errors from response.
            response = kwargs.get("response")

            if response is not None:
                try:
                    response_json = response.json()

                    self.errors = response_json.get("errors")
                except ValueError:
                    pass

        if self.errors is not None:
            args = (self.errors, *args)

        super().__init__(*args, **kwargs)



ORDWAY_APP_DOMAIN = "app.ordwaylabs.com"
ORDWAY_APP_URL = f"https://{ORDWAY_APP_DOMAIN}"

ORDWAY_STAGING_DOMAIN = "staging.ordwaylabs.com"
ORDWAY_STAGING_URL = f"https://{ORDWAY_STAGING_DOMAIN}"

API_ENDPOINT_BASE = f"{ORDWAY_APP_URL}/api"
STAGING_ENDPOINT_BASE = f"{ORDWAY_STAGING_URL}/api"

SUPPORTED_API_VERSIONS = ("1",)



from typing import TYPE_CHECKING, Optional, Dict, List
from logging import getLogger
from os import environ
from requests.exceptions import RequestException

from .session import session_factory
from .exceptions import OrdwayClientException
from .consts import SUPPORTED_API_VERSIONS
from .api import (
    Products,
    Invoices,
    Customers,
    Subscriptions,
    Payments,
    Credits,
    Refunds,
    Plans,
    Webhooks,
)

if TYPE_CHECKING:
    from requests import Session

logger = getLogger(__name__)


class OrdwayClient:
    SUPPORTED_API_VERSIONS = SUPPORTED_API_VERSIONS

    def __init__(
        self,
        email: str,
        api_key: str,
        company: str,
        user_token: str,
        api_version: str = "1",
        staging: bool = False,
        proxies: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        session: Optional["Session"] = None,
    ):
        self.email = email
        self.api_key = api_key
        self.company = company
        self.user_token = user_token

        self.session = session_factory(session)
        self.headers = headers
        self.proxies = proxies

        if headers is not None:
            self.session.headers.update(headers)
        if proxies is not None:
            self.session.proxies.update(proxies)

        self.api_version = api_version

        # Interfaces
        self.products = Products(self, staging=staging)
        self.customers = Customers(self, staging=staging)
        self.subscriptions = Subscriptions(self, staging=staging)
        self.invoices = Invoices(self, staging=staging)
        self.payments = Payments(self, staging=staging)
        self.credits = Credits(self, staging=staging)
        self.plans = Plans(self, staging=staging)
        self.refunds = Refunds(self, staging=staging)
        self.webhooks = Webhooks(self, staging=staging)

    @property
    def api_version(self):
        return self._api_version

    @api_version.setter
    def api_version(self, api_version):
        if api_version.startswith("v"):
            api_version = api_version[1:]

        self._api_version = api_version
        self._verify_api_version()

    def _verify_api_version(self) -> None:
        """ Verifies that the requested API version is supported by OrdwayClient. """

        version = self.api_version

        if not version in self.SUPPORTED_API_VERSIONS:
            raise OrdwayClientException(
                f'OrdwayClient does not currently support the API version "v{version}".'
            )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.session.close()

    @classmethod
    def from_env(cls) -> "OrdwayClient":
        """ Instantiates `OrdwayClient` using environment variables. """

        try:
            env_vars = {
                "email": environ["ORDWAY_EMAIL"],
                "api_key": environ["ORDWAY_API_KEY"],
                "company": environ["ORDWAY_COMPANY"],
                "user_token": environ["ORDWAY_USER_TOKEN"],
            }

            logger.debug(
                'Instantiating client from environment with email "%s" and company "%s".',
                env_vars["email"],
                env_vars["company"],
            )
        except KeyError:
            err_message = "Cannot instantiate `OrdwayClient` with `.from_env`. Must set all of the following env vars: `ORDWAY_EMAIL`, `ORDWAY_API_KEY`, `ORDWAY_COMPANY`, and `ORDWAY_USER_TOKEN`."

            logger.error(err_message)

            raise OrdwayClientException(err_message)

        api_version = environ.get("ORDWAY_API_VERSION")

        if api_version is not None:
            env_vars["api_version"] = api_version

        # MyPy complains about incompatible type. Argument 1 to "OrdwayClient" has incompatible type "**Dict[str, str]"; expected "Optional[Dict[str, str]]"
        return cls(**env_vars)  # type: ignore



from typing import TYPE_CHECKING, Optional, List, Dict, Any, Generator, Union, Tuple
from logging import getLogger
from requests.exceptions import RequestException
from ordway.consts import API_ENDPOINT_BASE, STAGING_ENDPOINT_BASE

from .exceptions import OrdwayAPIRequestException, OrdwayAPIException

if TYPE_CHECKING:
    from ordway.client import OrdwayClient

logger = getLogger(__name__)

_Response = Union[List[Dict[str, Any]], Dict[str, Any]]


class APIBase:
    collection: str

    def __init__(self, client: "OrdwayClient", staging: bool = False):
        self.client = client
        self.session = client.session
        self.staging = staging

    def _construct_headers(self) -> Dict[str, str]:
        """ Returns a dictionary of headers Ordway always expects for API requests. """

        return {
            "X-User-Company": self.client.company,
            "X-API-Key": self.client.api_key,
            "X-User-Token": self.client.user_token,
            "X-User-Email": self.client.email,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

    # TODO May want to make static, or separate.
    def _request(
        self,
        method: str,
        endpoint: str,
        params: Optional[Dict] = None,
        data: Optional[Dict] = None,
        json: Optional[Dict] = None,
    ) -> _Response:  # TODO Be more specific here
        if self.staging:
            base = STAGING_ENDPOINT_BASE
        else:
            base = API_ENDPOINT_BASE

        url = f"{base}/v{self.client.api_version}/{endpoint}"

        logger.debug(
            'Sending a request to Ordway endpoint "%s" with the following query params: %s',
            endpoint,
            params,
        )

        # Ensure any changes to client attrs are reflected in headers on request.
        self.session.headers.update(self._construct_headers())

        try:
            response = self.session.request(
                method=method, url=url, params=params, data=data, json=json
            )

            response.raise_for_status()

            return response.json()
        except RequestException as err:
            raise OrdwayAPIRequestException(
                str(err), request=err.request, response=err.response
            )
        except ValueError:
            raise OrdwayAPIRequestException(
                "Ordway returned HTTP success, but no valid JSON was present. Please report this as an issue on GitHub."
            )

    def _get_request(self, endpoint: str, params: Optional[Dict] = None) -> _Response:
        return self._request("GET", endpoint, params=params)

    def _post_request(
        self,
        endpoint: str,
        data: Optional[Dict] = None,
        json: Optional[Dict] = None,
        params: Optional[Dict] = None,
    ) -> _Response:
        if json is None and data is None:
            raise ValueError("Either `json` or `data` must be passed to post_request.")

        return self._request("POST", endpoint, json=json, data=data, params=params)


def _remove_order_from_sort(sort_str: str) -> Tuple[str, Optional[bool]]:
    sort_str = sort_str.lower()

    if sort_str.endswith(" asc"):
        return (sort_str[:-4], True)
    if sort_str.endswith(" desc"):
        return (sort_str[:-5], False)

    return (sort_str, None)


class ListAPIMixin(APIBase):
    """ Mixin for retrieving a collection of Ordway resources. """

    MAX_PAGE_SIZE = 50
    MAX_PAGES = 1000

    def list(
        self,
        page: int,
        size: int = 20,
        sort: str = "",
        filters: Optional[Dict] = None,
        ascending: bool = False,
    ) -> Generator[Dict[str, Any], None, None]:
        if size > self.MAX_PAGE_SIZE:
            logger.warning(
                'Maximum page size for "%s" is %s, setting `size` to maximum.',
                self.collection,
                self.MAX_PAGE_SIZE,
            )

            size = self.MAX_PAGE_SIZE

        filters = {} if filters is None else filters

        # Ordway appends order onto the end of the sort string. It's best
        # to ensure that if a user does, we remove it. Sice we're going to
        # append it later.
        sort, order = _remove_order_from_sort(sort)

        if order is not None:
            ascending = order

        response_json = self._get_request(
            self.collection,
            params={
                "size": size,
                "page": page,
                "sort": f"{sort} {'asc' if ascending else 'desc'}",
                **filters,
            },
        )

        if isinstance(response_json, dict):
            response_json = [response_json]

        # Mostly for consistency's sake.
        for result in response_json:
            yield result

        if len(response_json) < size:
            self._exhausted = True

    def all(
        self,
        size: int = 20,
        sort: str = "",
        filters: Optional[Dict] = None,
        ascending: bool = False,
        ignore_max_pages: bool = False,
    ) -> Generator[Dict[str, Any], None, None]:
        page = 1

        self._exhausted = False

        # Maybe separate into Paginator class? Could be needed elsewhere.
        while True:
            if not ignore_max_pages and page >= self.MAX_PAGES:
                logger.warning(
                    "Call to `.all()` has reached the maximum number of pages (%s). If this is desirable, please call with `ignore_max_pages` set to True.",
                    self.MAX_PAGES,
                )

                break

            yield from self.list(
                page=page,
                size=size,
                sort=sort,
                filters=filters,
                ascending=ascending,
            )

            if self._exhausted:
                break

            page += 1


class GetAPIMixin(APIBase):
    """ Mixin for retrieving a single Ordway resource. """

    def get(self, id: str) -> Dict[str, Any]:
        response_json = self._get_request(f"{self.collection}/{id}")

        if isinstance(response_json, list):
            if len(response_json) == 1:
                return response_json[0]
            else:
                raise OrdwayAPIException(
                    "Call to `.get_request` returned an unexpected JSON array. Please report this as an issue on GitHub."
                )

        return response_json


class CreateAPIMixin(APIBase):
    """ Mixin for creating a single Ordway resource. """

    def create(self, json: Optional[Dict], params: Optional[Dict] = None) -> _Response:
        return self._post_request(self.collection, json=json, data=None, params=params)


class UpdateAPIMixin(APIBase):
    pass


class DeleteAPIMixin(APIBase):
    """ Mixin for deleting a single Ordway resource. """

    def delete(self, id: str) -> _Response:
        return self._request("DELETE", f"{self.collection}/{id}")


__all__ = [
    "APIBase",
    "ListAPIMixin",
    "GetAPIMixin",
    "CreateAPIMixin",
    "UpdateAPIMixin",
    "DeleteAPIMixin",
]
